# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Define universal constants
"""
#: Gravitational constant :math:`G` in SI units (:math:`m^3 kg^{-1} s^{-2}`)
GRAVITATIONAL_CONST = 6.6743e-11

#: Vacuum magnetic permeability :math:`\mu_0` in SI units (:math:`NA^{-2}`)
VACUUM_MAGNETIC_PERMEABILITY = 1.25663706212e-06
